@@ -3,7 +3,7 @@
  * mapping-method.c - VFS modules for handling remapped files
  *
  * Copyright (C) 2002 Red Hat Inc,
- * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2005-2006 William Jon McCann <mccann@jhu.edu>
  *
  * The Gnome Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -24,6 +24,8 @@
  *          William Jon McCann <mccann@jhu.edu>
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -53,13 +55,6 @@
 
 MappingProtocolChannel *daemon_ioc = NULL;
 
-G_LOCK_DEFINE_STATIC (mapping_lock);
-
-static gint monitor_watch_id = 0;
-
-static void     monitor_stop_unlocked  (void);
-static gboolean monitor_setup (void);
-
 #define LAUNCH_DAEMON_TIMEOUT 2*1000
 
 #undef DEBUG_ENABLE
@@ -114,18 +109,24 @@ static void
 debug_print (const char *format, ...)
 {
 	va_list args;
+	char    buffer [1025];
 
 	va_start (args, format);
-	vfprintf ((debug_out ? debug_out : stderr), format, args);
-	vfprintf (stderr, format, args);
+	g_vsnprintf (buffer, 1024, format, args);
 	va_end (args);
-	if (debug_out)
+
+	fprintf ((debug_out ? debug_out : stderr),
+		 "[%p] %s",
+		 g_thread_self (), buffer);
+
+	if (debug_out) {
 		fflush (debug_out);
+	}
 }
 #endif
 
 static char *
-get_path_from_uri (GnomeVFSURI const *uri)
+get_path_from_uri (const GnomeVFSURI *uri)
 {
 	char *path;
 
@@ -162,149 +163,95 @@ get_uri (char *path)
 	return uri;
 }
 
-static GnomeVFSResult
-request_op_unlocked (gint32                operation,
-		     char                 *root,
-		     char                 *path1,
-		     char                 *path2,
-		     gboolean              option,
-		     void                 *userdata,
-		     MappingProtocolReply *reply)
+static void
+process_event (MappingProtocolMonitorEvent *event)
 {
-	int res;
+	VirtualFileMonitorHandle *handle;
+	gboolean                  cancelled;
+	GnomeVFSMonitorEventType  event_type;
 
-	DEBUG_PRINT ("request_op_unlocked: %d\n", operation);
+	if (!event->userdata) {
+		DEBUG_PRINT ("Received event with no userdata, skipping\n");
+		return;
+	}
 
-	res = mapping_protocol_request_encode (daemon_ioc,
-					       operation,
-					       root,
-					       path1,
-					       path2,
-					       option,
-					       userdata);
-	if (res != 0)
-		return GNOME_VFS_ERROR_IO;
+	handle = (VirtualFileMonitorHandle *)event->userdata;
+	cancelled = handle->cancelled;
+	event_type = event->type;
 
-	DEBUG_PRINT ("request_op_unlocked: reply_decode %d\n", operation);
-	res = mapping_protocol_reply_decode (daemon_ioc, reply);
-	if (res != 0)
-		return GNOME_VFS_ERROR_IO;
+	if (!cancelled) {
+		GnomeVFSURI *info_uri;
 
-	DEBUG_PRINT ("request_op_unlocked: DONE %d\n", reply->result);
+		/*
+		 * can send events with either a absolute or
+		 * relative (from the monitored URI) path, so check if
+		 * the filename starts with G_DIR_SEPARATOR.
+		 */
+		if (g_path_is_absolute (event->path)) {
+			char        *info_str;
+			GnomeVFSURI *tmp_uri;
 
-	return reply->result;
-}
+			info_str = g_strdup_printf ("%s://", handle->uri->method_string);
+			/*info_str = g_strdup ("burn://");*/
+			tmp_uri = gnome_vfs_uri_new (info_str);
+			info_uri = gnome_vfs_uri_append_path (tmp_uri, event->path);
+			gnome_vfs_uri_unref (tmp_uri);
+			g_free (info_str);
+		} else {
+			info_uri = gnome_vfs_uri_append_file_name (handle->uri, event->path);
+		}
 
-static int
-monitor_event_pending (MappingProtocolMonitorEvent *event)
-{
-	int res;
+		DEBUG_PRINT ("Returning event %d path %s handle %p\n",
+			     event_type,
+			     event->path,
+			     handle);
 
-	res = mapping_protocol_data_available (daemon_ioc);
-
-	if (res <= 0)
-		return -1;
-
-	res = mapping_protocol_monitor_event_decode (daemon_ioc, event);
-
-	return res;
+		/* This queues an idle, so there are no reentrancy issues */
+		gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *)handle,
+					    info_uri,
+					    event_type);
+		gnome_vfs_uri_unref (info_uri);
+	} else {
+		DEBUG_PRINT ("Cancelled event %d path %s handle %p\n",
+			     event_type,
+			     event->path,
+			     handle);
+	}
 }
 
 static gboolean
-monitor_do_iter_unlocked (void)
+request_op (gint32                   operation,
+	    char                    *root,
+	    char                    *path1,
+	    char                    *path2,
+	    gboolean                 option,
+	    void                    *userdata,
+	    MappingProtocolMessage **reply)
 {
-	MappingProtocolMonitorEvent event;
+	gboolean                res;
+	MappingProtocolMessage *message;
+	MappingProtocolRequest *request;
 
-	while (monitor_event_pending (&event) == 0) {
-		VirtualFileMonitorHandle *handle;
-		gboolean                  cancelled;
-		GnomeVFSMonitorEventType  event_type;
+	DEBUG_PRINT ("request_op: %d\n", operation);
 
-		if (!event.userdata) {
-			DEBUG_PRINT ("Received event with no userdata, skipping\n");
-			continue;
-		}
+	message = mapping_protocol_message_new_request ();
+	request = MAPPING_PROTOCOL_REQUEST (message);
 
-		handle = (VirtualFileMonitorHandle *)event.userdata;
-		cancelled = handle->cancelled;
-		event_type = event.type;
+	request->operation = operation;
+	request->root      = g_strdup (root);
+	request->path1     = g_strdup (path1);
+	request->path2     = g_strdup (path2);
+	request->option    = option;
+	request->userdata  = userdata;
 
-		if (!cancelled) {
-			GnomeVFSURI *info_uri;
+	res = mapping_protocol_channel_send_with_reply (daemon_ioc, message, reply);
 
-			/*
-			 * can send events with either a absolute or
-			 * relative (from the monitored URI) path, so check if
-			 * the filename starts with G_DIR_SEPARATOR.
-			 */
-			if (g_path_is_absolute (event.path)) {
-				char        *info_str;
-				GnomeVFSURI *tmp_uri;
-
-				info_str = g_strdup_printf ("%s://", handle->uri->method_string);
-				/*info_str = g_strdup ("burn://");*/
-				tmp_uri = gnome_vfs_uri_new (info_str);
-				info_uri = gnome_vfs_uri_append_path (tmp_uri, event.path);
-				gnome_vfs_uri_unref (tmp_uri);
-				g_free (info_str);
-			} else
-				info_uri = gnome_vfs_uri_append_file_name (handle->uri, event.path);
-
-			DEBUG_PRINT ("Returning event %d path %s handle %p\n",
-				     event_type,
-				     event.path,
-				     handle);
-
-			/* This queues an idle, so there are no reentrancy issues */
-			gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *)handle,
-						    info_uri,
-						    event_type);
-			gnome_vfs_uri_unref (info_uri);
-		} else {
-			DEBUG_PRINT ("Cancelled event %d path %s handle %p\n",
-				     event_type,
-				     event.path,
-				     handle);
-		}
+	if (! res) {
+		DEBUG_PRINT ("request_op: send ERROR\n");
+	} else {
+		DEBUG_PRINT ("request_op: send OK\n");
 	}
 
-	return TRUE;
-}
-
-static GnomeVFSResult
-request_op (gint32                operation,
-	    char                 *root,
-	    char                 *path1,
-	    char                 *path2,
-	    gboolean              option,
-	    void                 *userdata,
-	    MappingProtocolReply *reply)
-{
-	GnomeVFSResult res;
-	gboolean       monitor_active;
-
-	G_LOCK (mapping_lock);
-
-	monitor_active = (monitor_watch_id > 0);
-
-	/* disable the IO Channel watcher while reading reply */
-	monitor_stop_unlocked ();
-
-	monitor_do_iter_unlocked ();
-
-	res = request_op_unlocked (operation,
-				   root,
-				   path1,
-				   path2,
-				   option,
-				   userdata,
-				   reply);
-	G_UNLOCK (mapping_lock);
-
-	/* re-enable the IO Channel watcher after reading reply */
-	if (monitor_active)
-		monitor_setup ();
-
 	return res;
 }
 
@@ -312,13 +259,23 @@ static GnomeVFSResult
 remove_file_helper (char *method,
 		    char *path)
 {
-	MappingProtocolReply reply;
-	GnomeVFSResult       res;
+	MappingProtocolMessage *reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_REMOVE_FILE, method,
 			  path, NULL, FALSE, NULL, &reply);
-	mapping_protocol_reply_destroy (&reply);
-	return res;
+
+
+	if (! res) {
+		return GNOME_VFS_ERROR_IO;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+	mapping_protocol_message_unref (reply);
+
+	return ret;
 }
 
 static GnomeVFSResult
@@ -328,14 +285,15 @@ do_open (GnomeVFSMethod        *method,
 	 GnomeVFSOpenMode       mode,
 	 GnomeVFSContext       *context)
 {
-	GnomeVFSResult       res;
-	GnomeVFSURI         *file_uri = NULL;
-	GnomeVFSHandle      *file_handle;
-	VirtualFileHandle   *handle;
-	char                *path;
-	MappingProtocolReply reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	GnomeVFSURI            *file_uri = NULL;
+	GnomeVFSHandle         *file_handle;
+	VirtualFileHandle      *handle;
+	char                   *path;
+	MappingProtocolMessage *reply;
 
-	DEBUG_PRINT ("do_open: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_open: %s\n", uri->text);
 
 	*method_handle = NULL;
 	path = get_path_from_uri (uri);
@@ -343,6 +301,7 @@ do_open (GnomeVFSMethod        *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_GET_BACKING_FILE,
 			  uri->method_string,
 			  path,
@@ -352,24 +311,32 @@ do_open (GnomeVFSMethod        *method,
 			  &reply);
 	g_free (path);
 
-	if (res == GNOME_VFS_OK) {
-		file_uri = get_uri (reply.path);
-		res = gnome_vfs_open_uri_cancellable (&file_handle,
+	if (! res) {
+		return GNOME_VFS_ERROR_IO;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+
+	if (ret == GNOME_VFS_OK) {
+		file_uri = get_uri (MAPPING_PROTOCOL_REPLY (reply)->path);
+		ret = gnome_vfs_open_uri_cancellable (&file_handle,
 						      file_uri, mode, context);
 
-		if (res == GNOME_VFS_OK) {
+		if (ret == GNOME_VFS_OK) {
 			handle = g_new (VirtualFileHandle, 1);
 			handle->handle = file_handle;
-			handle->backing_file = g_strdup (reply.path);
+			handle->backing_file = g_strdup (MAPPING_PROTOCOL_REPLY (reply)->path);
 			*method_handle = (GnomeVFSMethodHandle *)handle;
 		}
 
 		gnome_vfs_uri_unref (file_uri);
 	}
 
-	mapping_protocol_reply_destroy (&reply);
+	if (reply) {
+		mapping_protocol_message_unref (reply);
+	}
 
-	return res;
+	return ret;
 }
 
 static GnomeVFSResult
@@ -381,15 +348,16 @@ do_create (GnomeVFSMethod        *method,
 	   guint                  perm,
 	   GnomeVFSContext       *context)
 {
-	GnomeVFSResult       res;
-	GnomeVFSURI         *file_uri;
-	char                *path;
-	MappingProtocolReply reply;
-	gboolean             newly_created;
-	GnomeVFSHandle      *file_handle;
-	VirtualFileHandle   *handle;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	GnomeVFSURI            *file_uri;
+	char                   *path;
+	MappingProtocolMessage *reply;
+	gboolean                newly_created;
+	GnomeVFSHandle         *file_handle;
+	VirtualFileHandle      *handle;
 
-	DEBUG_PRINT ("do_create: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_create: %s\n", uri->text);
 
 	*method_handle = NULL;
 	path = get_path_from_uri (uri);
@@ -397,40 +365,46 @@ do_create (GnomeVFSMethod        *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_CREATE_FILE,
 			  uri->method_string,
 			  path, NULL,
 			  exclusive,
 			  NULL,
 			  &reply);
-	newly_created = reply.option;
 
-	if (res == GNOME_VFS_OK) {
-		file_uri = get_uri (reply.path);
+	if (! res) {
+		return GNOME_VFS_ERROR_IO;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+	newly_created = MAPPING_PROTOCOL_REPLY (reply)->option;
+
+	if (ret == GNOME_VFS_OK) {
+		file_uri = get_uri (MAPPING_PROTOCOL_REPLY (reply)->path);
 		res = gnome_vfs_create_uri_cancellable
 			(&file_handle,
 			 file_uri, mode, exclusive,
 			 perm, context);
 		gnome_vfs_uri_unref (file_uri);
-		if (res == GNOME_VFS_OK) {
+		if (ret == GNOME_VFS_OK) {
 			handle = g_new (VirtualFileHandle, 1);
 			handle->handle = file_handle;
-			handle->backing_file = g_strdup (reply.path);
+			handle->backing_file = g_strdup (MAPPING_PROTOCOL_REPLY (reply)->path);
 			*method_handle = (GnomeVFSMethodHandle *)handle;
 		}
-
 	}
 
-	mapping_protocol_reply_destroy (&reply);
+	mapping_protocol_message_unref (reply);
 
-	if (res != GNOME_VFS_OK && newly_created) {
+	if (ret != GNOME_VFS_OK && newly_created) {
 			/* TODO: Remove the file that was created, since we failed */
 			/* virtual_unlink (dir, file);
 			 * virtual_node_free (file, FALSE);
 			 */
 	}
 
-	return res;
+	return ret;
 }
 
 static GnomeVFSResult
@@ -529,12 +503,13 @@ do_open_directory (GnomeVFSMethod         *method,
 		   GnomeVFSFileInfoOptions options,
 		   GnomeVFSContext        *context)
 {
-	char                *path;
-	VirtualDirHandle    *handle;
-	MappingProtocolReply reply;
-	GnomeVFSResult       res;
+	char                   *path;
+	VirtualDirHandle       *handle;
+	MappingProtocolMessage *reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
 
-	DEBUG_PRINT ("do_open_directory: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_open_directory: %s\n", uri->text);
 
 	g_return_val_if_fail (uri != NULL, GNOME_VFS_ERROR_BAD_PARAMETERS);
 
@@ -543,6 +518,7 @@ do_open_directory (GnomeVFSMethod         *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_LIST_DIR,
 			  uri->method_string,
 			  path, NULL,
@@ -550,14 +526,27 @@ do_open_directory (GnomeVFSMethod         *method,
 			  NULL,
 			  &reply);
 
-	if (res == GNOME_VFS_OK) {
+	if (! res) {
+		return GNOME_VFS_ERROR_IO;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+
+	if (ret == GNOME_VFS_OK) {
+		int i;
+
 		handle = g_new (VirtualDirHandle, 1);
 
 		handle->pos = 0;
 		handle->dirname = path;
-		handle->listing = reply.strings;
-		g_assert ((reply.n_strings % 2) == 0);
-		handle->n_items = reply.n_strings / 2;
+
+		g_assert ((MAPPING_PROTOCOL_REPLY (reply)->n_strings % 2) == 0);
+		handle->n_items = MAPPING_PROTOCOL_REPLY (reply)->n_strings / 2;
+		handle->listing = g_new0 (char *, MAPPING_PROTOCOL_REPLY (reply)->n_strings);
+		for (i = 0; i < MAPPING_PROTOCOL_REPLY (reply)->n_strings; i++) {
+			handle->listing[i] = g_strdup (MAPPING_PROTOCOL_REPLY (reply)->strings[i]);
+		}
+
 		handle->root = g_strdup (uri->method_string);
 		handle->options = options;
 
@@ -566,8 +555,9 @@ do_open_directory (GnomeVFSMethod         *method,
 		g_free (path);
 	}
 
-	mapping_protocol_reply_destroy (&reply);
-	return res;
+	mapping_protocol_message_unref (reply);
+
+	return ret;
 }
 
 static GnomeVFSResult
@@ -662,18 +652,20 @@ do_get_file_info (GnomeVFSMethod         *method,
 		  GnomeVFSFileInfoOptions options,
 		  GnomeVFSContext        *context)
 {
-	GnomeVFSResult       res;
-	GnomeVFSURI         *file_uri = NULL;
-	char                *path;
-	MappingProtocolReply reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	GnomeVFSURI            *file_uri = NULL;
+	char                   *path;
+	MappingProtocolMessage *reply;
 
-	DEBUG_PRINT ("do_get_file_info: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_get_file_info: %s\n", uri->text);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_GET_BACKING_FILE,
 			  uri->method_string,
 			  path,
@@ -681,26 +673,35 @@ do_get_file_info (GnomeVFSMethod         *method,
 			  FALSE,
 			  NULL,
 			  &reply);
-	DEBUG_PRINT ("do_get_file_info: file %s result %d\n", path, res);
 
-	if (res == GNOME_VFS_ERROR_IS_DIRECTORY) {
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+		DEBUG_PRINT ("do_get_file_info: file %s IO ERROR\n", path);
+		goto out;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+	DEBUG_PRINT ("do_get_file_info: file %s result %s\n", path, gnome_vfs_result_to_string (ret));
+
+	if (ret == GNOME_VFS_ERROR_IS_DIRECTORY) {
 		file_info->name = g_path_get_basename (path);
 		fill_in_directory_info (file_info);
-		res = GNOME_VFS_OK;
-	} else if (res == GNOME_VFS_OK) {
-		file_uri = get_uri (reply.path);
-		res = gnome_vfs_get_file_info_uri_cancellable
-			(file_uri, file_info, options, context);
+		ret = GNOME_VFS_OK;
+	} else if (ret == GNOME_VFS_OK) {
+		file_uri = get_uri (MAPPING_PROTOCOL_REPLY (reply)->path);
+		ret = gnome_vfs_get_file_info_uri_cancellable (file_uri, file_info, options, context);
 		gnome_vfs_uri_unref (file_uri);
 
 		g_free (file_info->name);
 		file_info->name = g_path_get_basename (path);
 	}
 
-	mapping_protocol_reply_destroy (&reply);
+	mapping_protocol_message_unref (reply);
+
+ out:
 	g_free (path);
 
-	return res;
+	return ret;
 }
 
 static GnomeVFSResult
@@ -711,10 +712,10 @@ do_get_file_info_from_handle (GnomeVFSMethod         *method,
 			      GnomeVFSContext        *context)
 {
 	GnomeVFSResult res;
+
 	DEBUG_PRINT ("do_get_file_info_from_handle: %p\n", method_handle);
-	res = gnome_vfs_get_file_info_from_handle_cancellable
-		(((VirtualFileHandle *)method_handle)->handle,
-		 file_info, options, context);
+	res = gnome_vfs_get_file_info_from_handle_cancellable (((VirtualFileHandle *)method_handle)->handle,
+							       file_info, options, context);
 	/* TODO: Need to fill out the real name here. Need to wrap method_handle */
 	return res;
 }
@@ -725,27 +726,36 @@ do_make_directory (GnomeVFSMethod  *method,
 		   guint            perm,
 		   GnomeVFSContext *context)
 {
-	GnomeVFSResult       res;
-	char                *path;
-	MappingProtocolReply reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	char                   *path;
+	MappingProtocolMessage *reply;
 
-	DEBUG_PRINT ("do_make_directory: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_make_directory: %s\n", uri->text);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_CREATE_DIR,
 			  uri->method_string,
 			  path,
 			  NULL, FALSE,
 			  NULL,
 			  &reply);
-	g_free (path);
-	mapping_protocol_reply_destroy (&reply);
 
-	return res;
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+	} else {
+		ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+		mapping_protocol_message_unref (reply);
+	}
+
+	g_free (path);
+
+	return ret;
 }
 
 static GnomeVFSResult
@@ -753,27 +763,38 @@ do_remove_directory (GnomeVFSMethod  *method,
 		     GnomeVFSURI     *uri,
 		     GnomeVFSContext *context)
 {
-	GnomeVFSResult       res;
-	char                *path;
-	MappingProtocolReply reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	char                   *path;
+	MappingProtocolMessage *reply;
 
-	DEBUG_PRINT ("do_remove_directory: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_remove_directory: %s\n", uri->text);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_REMOVE_DIR,
 			  uri->method_string,
 			  path,
 			  NULL, FALSE,
 			  NULL,
 			  &reply);
-	g_free (path);
-	mapping_protocol_reply_destroy (&reply);
 
-	return res;
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+		goto out;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+
+	mapping_protocol_message_unref (reply);
+ out:
+	g_free (path);
+
+	return ret;
 }
 
 static GnomeVFSResult
@@ -784,7 +805,7 @@ do_unlink (GnomeVFSMethod  *method,
 	GnomeVFSResult res;
 	char          *path;
 
-	DEBUG_PRINT ("do_remove_directory: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_remove_directory: %s\n", uri->text);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
@@ -804,12 +825,13 @@ do_move (GnomeVFSMethod  *method,
 	 gboolean         force_replace,
 	 GnomeVFSContext *context)
 {
-	GnomeVFSResult       res;
-	MappingProtocolReply reply;
-	char                *old_path;
-	char                *new_path;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	MappingProtocolMessage *reply;
+	char                   *old_path;
+	char                   *new_path;
 
-	DEBUG_PRINT ("do_move (%s, %s)\n", gnome_vfs_uri_to_string (old_uri, 0), gnome_vfs_uri_to_string (new_uri, 0));
+	DEBUG_PRINT ("do_move (%s, %s)\n", old_uri->text, new_uri->text);
 
 	if (strcmp (new_uri->method_string,
 		    old_uri->method_string) != 0) {
@@ -827,6 +849,7 @@ do_move (GnomeVFSMethod  *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_MOVE_FILE,
 			  old_uri->method_string,
 			  old_path,
@@ -834,11 +857,21 @@ do_move (GnomeVFSMethod  *method,
 			  FALSE,
 			  NULL,
 			  &reply);
-	mapping_protocol_reply_destroy (&reply);
+
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+		goto out;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+
+	mapping_protocol_message_unref (reply);
+
+ out:
 	g_free (old_path);
 	g_free (new_path);
 
-	return res;
+	return ret;
 }
 
 static gboolean
@@ -873,12 +906,13 @@ do_set_file_info (GnomeVFSMethod         *method,
 		  GnomeVFSSetFileInfoMask mask,
 		  GnomeVFSContext        *context)
 {
-	char                *full_name;
-	GnomeVFSResult       res;
-	MappingProtocolReply reply;
-	GnomeVFSURI         *file_uri = NULL;
+	char                   *full_name;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	MappingProtocolMessage *reply;
+	GnomeVFSURI            *file_uri = NULL;
 
-	DEBUG_PRINT ("do_set_file_info: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_set_file_info: %s\n", uri->text);
 
 	full_name = get_path_from_uri (uri);
 	if (full_name == NULL) {
@@ -905,6 +939,7 @@ do_set_file_info (GnomeVFSMethod         *method,
 			new_name = g_strconcat (dir, info->name, NULL);
 		}
 
+		reply = NULL;
 		res = request_op (MAPPING_PROTOCOL_OP_MOVE_FILE,
 				  uri->method_string,
 				  full_name,
@@ -912,20 +947,27 @@ do_set_file_info (GnomeVFSMethod         *method,
 				  FALSE,
 				  NULL,
 				  &reply);
-		mapping_protocol_reply_destroy (&reply);
+
+		if (! res) {
+			ret = GNOME_VFS_ERROR_IO;
+		} else {
+			ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+			mapping_protocol_message_unref (reply);
+		}
 
 		g_free (dir);
 		g_free (full_name);
 		full_name = new_name;
 
-		if (res != GNOME_VFS_OK) {
+		if (ret != GNOME_VFS_OK) {
 			g_free (full_name);
-			return res;
+			return ret;
 		}
 		mask = mask & ~GNOME_VFS_SET_FILE_INFO_NAME;
 	}
 
 	if (mask != 0) {
+		reply = NULL;
 		res = request_op (MAPPING_PROTOCOL_OP_GET_BACKING_FILE,
 				  uri->method_string,
 				  full_name,
@@ -933,16 +975,25 @@ do_set_file_info (GnomeVFSMethod         *method,
 				  TRUE,
 				  NULL,
 				  &reply);
+
 		g_free (full_name);
-		if (res != GNOME_VFS_OK) {
-			mapping_protocol_reply_destroy (&reply);
-			return res;
+
+		if (! res) {
+			return GNOME_VFS_ERROR_IO;
 		}
 
-		file_uri = get_uri (reply.path);
-		mapping_protocol_reply_destroy (&reply);
+		ret = MAPPING_PROTOCOL_REPLY (reply)->result;
 
-		res = gnome_vfs_set_file_info_cancellable (file_uri,
+		if (ret != GNOME_VFS_OK) {
+			mapping_protocol_message_unref (reply);
+			return ret;
+		}
+
+		file_uri = get_uri (MAPPING_PROTOCOL_REPLY (reply)->path);
+
+		mapping_protocol_message_unref (reply);
+
+		ret = gnome_vfs_set_file_info_cancellable (file_uri,
 							   info, mask, context);
 		gnome_vfs_uri_unref (file_uri);
 	}
@@ -956,18 +1007,20 @@ do_truncate (GnomeVFSMethod   *method,
 	     GnomeVFSFileSize  where,
 	     GnomeVFSContext  *context)
 {
-	GnomeVFSResult       res;
-	GnomeVFSURI         *file_uri = NULL;
-	char                *path;
-	MappingProtocolReply reply;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	GnomeVFSURI            *file_uri = NULL;
+	char                   *path;
+	MappingProtocolMessage *reply;
 
-	DEBUG_PRINT ("do_truncate: %s\n", gnome_vfs_uri_to_string (uri, 0));
+	DEBUG_PRINT ("do_truncate: %s\n", uri->text);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_GET_BACKING_FILE,
 			  uri->method_string,
 			  path,
@@ -976,14 +1029,20 @@ do_truncate (GnomeVFSMethod   *method,
 			  NULL,
 			  &reply);
 
-	if (res == GNOME_VFS_OK) {
-		file_uri = get_uri (reply.path);
+	if (! res) {
+		return GNOME_VFS_ERROR_IO;
+	}
+
+	ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+	if (ret == GNOME_VFS_OK) {
+		file_uri = get_uri (MAPPING_PROTOCOL_REPLY (reply)->path);
 		res = gnome_vfs_truncate_uri_cancellable (file_uri, where, context);
 		gnome_vfs_uri_unref (file_uri);
 	}
-	mapping_protocol_reply_destroy (&reply);
 
-	return res;
+	mapping_protocol_message_unref (reply);
+
+	return ret;
 }
 
 static GnomeVFSResult
@@ -992,12 +1051,13 @@ do_create_symbolic_link (GnomeVFSMethod  *method,
 			 const char      *target_reference,
 			 GnomeVFSContext *context)
 {
-	GnomeVFSResult       res;
-	MappingProtocolReply reply;
-	char                *path;
-	char                *target_path;
+	GnomeVFSResult          ret;
+	gboolean                res;
+	MappingProtocolMessage *reply;
+	char                   *path;
+	char                   *target_path;
 
-	DEBUG_PRINT ("do_create_symbolic_link: %s -> %s\n", gnome_vfs_uri_to_string (uri, 0), target_reference);
+	DEBUG_PRINT ("do_create_symbolic_link: %s -> %s\n", uri->text, target_reference);
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
@@ -1011,6 +1071,7 @@ do_create_symbolic_link (GnomeVFSMethod  *method,
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_CREATE_LINK,
 			  uri->method_string,
 			  path,
@@ -1018,56 +1079,18 @@ do_create_symbolic_link (GnomeVFSMethod  *method,
 			  FALSE,
 			  NULL,
 			  &reply);
-	mapping_protocol_reply_destroy (&reply);
+
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+	} else {
+		ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+		mapping_protocol_message_unref (reply);
+	}
+
 	g_free (target_path);
 	g_free (path);
 
-	return res;
-}
-
-static gboolean
-monitor_callback (GIOChannel  *source,
-		  GIOCondition condition,
-		  gpointer     data)
-{
-	gboolean res = TRUE;
-
-	if (condition & G_IO_IN) {
-		G_LOCK (mapping_lock);
-		res = monitor_do_iter_unlocked ();
-		G_UNLOCK (mapping_lock);
-	}
-
-	if (condition & G_IO_HUP) {
-		res = FALSE;
-	}
-
-	return res;
-}
-
-static void
-monitor_stop_unlocked (void)
-{
-	if (monitor_watch_id > 0) {
-		g_source_remove (monitor_watch_id);
-		monitor_watch_id = 0;
-	}
-}
-
-static gboolean
-monitor_setup (void)
-{
-	G_LOCK (mapping_lock);
-
-	if (monitor_watch_id == 0) {
-		monitor_watch_id = g_io_add_watch (daemon_ioc->iochannel,
-						   G_IO_IN | G_IO_HUP | G_IO_ERR,
-						   monitor_callback, NULL);
-	}
-
-	G_UNLOCK (mapping_lock);
-
-	return TRUE;
+	return ret;
 }
 
 static GnomeVFSResult
@@ -1077,13 +1100,10 @@ do_monitor_add (GnomeVFSMethod        *method,
 		GnomeVFSMonitorType    monitor_type)
 {
 	VirtualFileMonitorHandle *monitor;
-	MappingProtocolReply      reply;
+	MappingProtocolMessage   *reply;
 	char                     *path;
-	int                       res;
-
-	if (!monitor_setup ()) {
-		return GNOME_VFS_ERROR_NOT_SUPPORTED;
-	}
+	GnomeVFSResult            ret;
+	gboolean                  res;
 
 	path = get_path_from_uri (uri);
 	if (path == NULL) {
@@ -1096,6 +1116,7 @@ do_monitor_add (GnomeVFSMethod        *method,
 
 	DEBUG_PRINT ("do_monitor_add: %p path: %s\n", monitor, path);
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_MONITOR_ADD,
 			  monitor->uri->method_string,
 			  path,
@@ -1104,6 +1125,13 @@ do_monitor_add (GnomeVFSMethod        *method,
 			  monitor,
 			  &reply);
 
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+	} else {
+		ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+		mapping_protocol_message_unref (reply);
+	}
+
 	*method_handle_return = (GnomeVFSMethodHandle *)monitor;
 
 	g_free (path);
@@ -1117,8 +1145,9 @@ static GnomeVFSResult
 do_monitor_cancel (GnomeVFSMethod       *method,
 		   GnomeVFSMethodHandle *method_handle)
 {
-	MappingProtocolReply      reply;
-	int                       res;
+	MappingProtocolMessage   *reply;
+	GnomeVFSResult            ret;
+	gboolean                  res;
 	char                     *path;
 	VirtualFileMonitorHandle *monitor;
 
@@ -1126,10 +1155,6 @@ do_monitor_cancel (GnomeVFSMethod       *method,
 
 	DEBUG_PRINT ("do_monitor_cancel: %p\n", monitor);
 
-	if (!monitor_setup ()) {
-		return GNOME_VFS_ERROR_NOT_SUPPORTED;
-	}
-
 	if (monitor->cancelled)
 		return GNOME_VFS_OK;
 
@@ -1140,6 +1165,7 @@ do_monitor_cancel (GnomeVFSMethod       *method,
 
 	monitor->cancelled = TRUE;
 
+	reply = NULL;
 	res = request_op (MAPPING_PROTOCOL_OP_MONITOR_CANCEL,
 			  monitor->uri->method_string,
 			  path,
@@ -1148,6 +1174,13 @@ do_monitor_cancel (GnomeVFSMethod       *method,
 			  monitor,
 			  &reply);
 
+	if (! res) {
+		ret = GNOME_VFS_ERROR_IO;
+	} else {
+		ret = MAPPING_PROTOCOL_REPLY (reply)->result;
+		mapping_protocol_message_unref (reply);
+	}
+
 	gnome_vfs_uri_unref (monitor->uri);
 
 	g_free (monitor);
@@ -1282,6 +1315,30 @@ launch_daemon (void)
 	return TRUE;
 }
 
+static void
+handle_message (MappingProtocolChannel *channel,
+		MappingProtocolMessage *message,
+		gpointer                data)
+{
+
+	if (message == NULL) {
+		/* connection error */
+		return;
+	}
+
+	switch (message->type) {
+	case MAPPING_PROTOCOL_MESSAGE_MONITOR_EVENT:
+		process_event (MAPPING_PROTOCOL_MONITOR_EVENT (message));
+		break;
+	case MAPPING_PROTOCOL_MESSAGE_REPLY:
+		break;
+	case MAPPING_PROTOCOL_MESSAGE_REQUEST:
+		break;
+	default:
+		break;
+	}
+}
+
 GnomeVFSMethod *
 vfs_module_init (const char *method_name,
 		 const char *args)
@@ -1318,6 +1375,7 @@ vfs_module_init (const char *method_name,
         }
 
 	daemon_ioc = mapping_protocol_channel_new (daemon_fd);
+	mapping_protocol_channel_set_message_handler (daemon_ioc, handle_message, NULL);
 
 #ifdef DEBUG_ENABLE
 	debug_init ();
@@ -1329,8 +1387,6 @@ vfs_module_init (const char *method_name,
 void
 vfs_module_shutdown (GnomeVFSMethod *method)
 {
-	monitor_stop_unlocked ();
-
 	mapping_protocol_channel_unref (daemon_ioc);
 
 #ifdef DEBUG_ENABLE
